# fitting parameters of the empirical p-value distribution
# p: vector of observed p-values
# trunc: the number of outliers to truncate, from the lower and the upper tail
#
# nc, df, sc should be either a constant or a vector holding lower, starting,
# and upper values for noncentrality, degrees of freedom, and scale
#
# Example: fit the scale between 1 and 2, while holding noncentrality=0, df=1,
# and truncating 10 extreme p-values from both ends
#
#   > source("fit3par.r")
#   > nc<-0; df<-1; sc<-c(1,1.5,2)
#   > g <- Fit3par(p, nc, df, sc, 10)
#
# Now adjust p-values using the fitted parameter
#
#   > a <- array(0,n <- length(p))
#   > for(i in 1:n) a[i] <- pdist(p[i], g$nc, g$df, g$sc)
#

# p-value (p) non-central (nc) scaled (sc) chi-square distribution function
pdist <- function(p, nc, df, sc) {
  pchisq(qchisq(p, df,lower.tail=FALSE)/sc, df, ncp=nc, lower.tail=FALSE)
}

# Generating random p-values from a non-central, scaled chi-square
Rpns <- function(n, nc, df, sc) { 
    1 - pchisq(sc*qchisq(runif(n), df, ncp=nc), df)
}

# in this version, trunc is the p-value interval for which the estimation is done
# e.g. c(1e6, 0.05)
Fit3par.interval <- function(p, nc, df, sc, trunc=c(0,1)) {
  p <- sort(p)
  v <- seq(from=1/length(p), to=1, by=1/length(p))
  h <- ecdf(p)(v)
  v <- subset(v, p>trunc[1] & p<trunc[2])
  h <- subset(h, p>trunc[1] & p<trunc[2])

  if(length(df) == 1 && length(nc) == 1) {
    cat("fit just the scale\n");
    ofn <- function(x) { sum((h - pdist(v, nc, df, x))^2) }
    gf <- optim(sc[2], ofn,  method="L-BFGS-B", lower=sc[1], upper=sc[3])
    gf$sc <- gf$par; gf$nc <- nc; gf$df <- df
  } else if(length(df) == 1 && length(sc) == 1) {
    cat("fit just the noncentrality\n")
    ofn <- function(x) { sum((h - pdist(v, x, df, sc))^2) }
    gf <- optim(nc[2], ofn,  method="L-BFGS-B", lower=nc[1], upper=nc[3])
    gf$sc <- sc; gf$nc <- gf$par; gf$df <- df
  } else if(length(nc) == 1 && length(sc) == 1) {
    cat("fit just the df\n")
    ofn <- function(x) { sum((h - pdist(v, nc, x, sc))^2) }
    gf <- optim(df[2], ofn,  method="L-BFGS-B", lower=df[1], upper=df[3])
    gf$sc <- sc; gf$nc <- nc; gf$df <- gf$par
  } else if(length(df) == 1 && length(nc) == 3 && length(sc) == 3) {
    cat("fit nonc and scale\n")
    ofn <- function(x) { sum((h - pdist(v, x[1], df, x[2]))^2) }
    gf <- optim(c(nc[2], sc[2]), ofn,  method="L-BFGS-B",
                lower=c(nc[1],sc[1]), upper=c(nc[3],sc[3]))
    gf$nc <- gf$par[1]; gf$sc <- gf$par[2]; gf$df <- df
  } else if (length(sc) == 1 && length(df) == 3 && length(nc) == 3) {
    cat("fit nonc and df\n")
    ofn <- function(x) { sum((h - pdist(v, x[1], x[2], sc))^2) }
    gf <- optim(c(nc[2], df[2]), ofn,  method="L-BFGS-B",
                lower=c(nc[1],df[1]), upper=c(nc[3],df[3]))
    gf$nc <- gf$par[1]; gf$df <- gf$par[2]; gf$sc <- sc
  } else if (length(nc) == 1 && length(df) == 3 && length(sc) == 3) {
    cat("fit df and scale\n")
    ofn <- function(x) { sum((h - pdist(v, nc, x[1], x[2]))^2) }
    gf <- optim(c(df[2], sc[2]), ofn,  method="L-BFGS-B",
                lower=c(df[1],sc[1]), upper=c(df[3],sc[3]))
    gf$df <- gf$par[1]; gf$sc <- gf$par[2]; gf$nc <- nc
  } else {
    cat("fit all three parameters\n")
    ofn <- function(x) { sum((h - pdist(v, x[1], x[2], x[3]))^2) }
    gf <- optim(c(nc[2], df[2], sc[2]), ofn, method="L-BFGS-B",
                lower=c(nc[1],df[1],sc[1]), upper=c(nc[3],df[3],sc[3]))
    gf$nc <- gf$par[1]; gf$df <- gf$par[2]; gf$sc <- gf$par[3]
  }
  return (gf)
}
